"""
DataDog Utility methods.
"""
import time
from datadog_api_client import ApiClient, Configuration
from datadog_api_client.v2.api.metrics_api import MetricsApi
from datadog_api_client.v2.model.metric_intake_type import MetricIntakeType
from datadog_api_client.v2.model.metric_payload import MetricPayload
from datadog_api_client.v2.model.metric_point import MetricPoint
from datadog_api_client.v2.model.metric_series import MetricSeries


def submit_dataframe(
    host, api_key, app_key, metric_name, dataframe, datapoint_column,
    tag_columns=[], batch_size=500
):
    """
    Submit a pandas DataFrame to DataDog.

    Args:
        host (str): The DataDog API host.
        api_key (str): Your DataDog API key.
        app_key (str): Your DataDog application key.
        metric_name (str): The name of the metric to submit.
        dataframe (pandas.DataFrame): The DataFrame containing data to submit.
        datapoint_column (str): DataFrame column containing datapoints.
        tag_columns (list, optional): List of column names to be used as tags.
          Default is an empty list.
        batch_size (int, optional): Number of rows to submit in each batch.
          Default is 500.

    Returns:
        dict: A dictionary containing any errors encountered during submission.
          Format: {"errors": ["error_message_1", "error_message_2", ...]}.
    """
    responses = {"errors": []}
    # Setup configuration
    configuration = Configuration(host=host)
    configuration.api_key['apiKeyAuth'] = api_key
    configuration.api_key['appKeyAuth'] = app_key
    # Iterate in batches by provided batch size
    for start, end in [
        (i, min(i + batch_size, dataframe.shape[0]))
        for i in range(0, dataframe.shape[0], batch_size)
    ]:
        # Submit metrics to DataDog
        with ApiClient(configuration) as api_client:
            api_instance = MetricsApi(api_client)
            response = api_instance.submit_metrics(
                body=MetricPayload([
                    MetricSeries(
                        metric=metric_name,
                        type=MetricIntakeType.COUNT,
                        points=[MetricPoint(
                            timestamp=int(time.time()),
                            value=int(row[datapoint_column]))],
                        tags=[f"{t}:{row[t]}" for t in tag_columns])
                    for index, row in dataframe.loc[start:end].iterrows()]))

            responses["errors"] += response["errors"]

    return responses



""" This module will contain any query related functionality """
import os
import logging
import pymysql
from pymysql.constants import CLIENT
import psycopg2
import pandas as pd

# Setup s3 keys
S3_ACCESS_KEY = os.getenv('S3_ACCESS_KEY')
S3_SECRET_KEY = os.getenv('S3_SECRET_KEY')


def connect_to_db(driver, driver_args):
    """
        This method will establish a connection to the Redshift database.

        Args:
            driver (): database driver module
            driver_args (dict): map of database connection args

        Returns:
            db connnection, connection cursor, and success bool
    """
    connection_success = True
    try:
        if driver.__name__ in ('pymysql', 'psycopg2'):
            conn = driver.connect(**driver_args)
            cur = conn.cursor()
        elif driver.__name__ == "sqlalchemy":
            # Create the engine object
            engine = driver.create_engine(
                "mysql+pymysql://{}:{}@{}:{}/{}".format(
                    driver_args['user'], driver_args['password'],
                    driver_args['host'], driver_args['port'],
                    driver_args['database']))
            # Create a connection
            conn = engine.connect()
            cur = None
        else:
            logging.error("Invalid driver provided.")
            raise Exception
        logging.info("Database connection successful.")
    except Exception:
        conn = None
        cur = None
        connection_success = False
        logging.error("Database connection failure.", exc_info=True)

    return conn, cur, connection_success


def execute_query(conn, cur, query, message=None):
    """
        This method will safely execute the given query and log the result.

        Args:
            conn (psycopg2.extensions.connection): db connection
            cur (psycopg2.extensions.cursor): connection cursor
            query (str): sql query string
            message (str): logging message corresponging to query

        Returns:
            query success boolean
    """
    # Try to execute query given cursor and connection
    success = True
    try:
        cur.execute(query)
        conn.commit()
        if message:
            logging.info(message + " success.")
    # Catch exception and log error
    except Exception:
        success = False
        if message:
            logging.error(message + " failed.", exc_info=True)

    return success


def delete(
        conn, cur, table_name, load_location, delete_all=False, days=None,
        column_header=None):
    """
        This method will delete and optimize a table provided inputs.

        Args:
            conn (pymysql.connections.Connection): database connection
            table_name (str): name of database table
            load_location (str): database type
            delete_all (bool): whether to delete all rows
            days (int): number of days out to delete
            column_header (str): header for day filter

        Returns:
            deletion success boolean
    """
    delete_success = False
    # Construct delete query string
    delete_query = f"DELETE FROM {table_name}"
    if delete_all:
        delete_query += ";"
        msg = f"Deletion of all rows from {table_name}."
    elif ((days is not None) and (column_header is not None)):
        delete_query += (
            f" WHERE DATE({column_header}) <= "
            f"CURDATE() - INTERVAL {days} DAY;")
        msg = f"Deletion of rows beyond {days} days from {table_name}."
    else:
        logging.error(
            "Either provide delete_all True or both days and column_header")
        return delete_success
    # TODO: Change this to a keyword arg and remove load location
    # Add table optimization if load location is Aurora
    if "aurora" in load_location.lower():
        delete_query += f"\nOPTIMIZE TABLE {table_name};"

    return execute_query(conn, cur, delete_query, msg)


def copy_from_s3(
        load_location, driver_args, table_name, s3_location,
        file_format="csv", separator=",", header=True, delete_info={},
        replace=False, header_list=[]):
    """
        This method will allow you to copy data of different formats from s3
        to a database or data warehouse like AWS Aurora or AWS Redshift.

        Args:
            load_location (str): database type
            driver_args (string): database connection arguments
            table_name (string): database table name
            s3_location (string): prefix of dataset on s3
            separator (string): csv separator
            header_list (list): list of table headers in-order

        Returns:
            success boolean
    """
    load_success = False
    # Setup query
    if "aurora" in load_location.lower():
        driver = pymysql
        # Set ignore row string given header boolean
        ignore_rows = "\n    IGNORE 1 ROWS" if header else ""
        replace_rows = "REPLACE " if replace else ""
        # Setup list of headers if applicable
        header_info = ""
        if header_list:
            header_info = " " + str(tuple(f"`{i}`" for i in header_list))
        # Fill parameters
        query = f"""
            LOAD DATA FROM S3 PREFIX '{s3_location}'
            {replace_rows}INTO TABLE `{table_name}`
            FIELDS TERMINATED BY '{separator}'{ignore_rows}{header_info};
        """
    elif "redshift" in load_location.lower():
        driver = psycopg2
        # Setup conversion parameters for csv
        # TODO: implement header_info argument into redshift query format
        ignore_header = " IGNOREHEADER 1 " if header else " "
        conversion_params = (
            f"DELIMITER '{separator}'{ignore_header}MAXERROR AS 10000;")
        if file_format == "parquet":
            conversion_params = "FORMAT AS PARQUET;"
        # Fill parameters
        query = f"""
            COPY {table_name}
            FROM '{s3_location}'
            ACCESS_KEY_ID '{S3_ACCESS_KEY}'
            SECRET_ACCESS_KEY '{S3_SECRET_KEY}'
            {conversion_params}
        """
    # Establish database connection
    conn, cur, success = connect_to_db(driver, driver_args)
    # If connection is valid, load data
    if success:
        # Execute delete query if applicable
        delete_success = True
        if delete_info:
            delete_success = delete(
                conn, cur, table_name, load_location, **delete_info)
        # Copy data from s3 to Database / Data Warehouse
        load_success = execute_query(
            conn, cur, query, f"Loading of {s3_location} into {table_name}")
        # Close connection
        conn.close()

    return load_success


def collect_query_result(conn, query, message):
    """
        This method will safely collect the result of a query and return it as
        a pandas DataFrame.

        Args:
            conn (psycopg2.extensions.connection): db connection
            query (str): sql query string
            message (str): logging message corresponging to query

        Returns:
            query result and success boolean
    """
    success = True
    try:
        result = pd.read_sql(query, con=conn)
        logging.info(message + " success.")
    except Exception:
        success = False
        result = pd.DataFrame()
        logging.error(message + " failed.", exc_info=True)

    return result, success


def check_table_existence(conn, table_name):
    """
        This method will check the database schema for the provided table.

        Args:
            conn (pymysql.connections.Connection): database connection
            table_name (str): database table name

        Returns:
            existence boolean
    """
    table_exists = False
    # Check whether date table exists
    stmt = f"SHOW TABLES LIKE '{table_name}'"
    cur = conn.cursor()
    cur.execute(stmt)
    result = cur.fetchone()
    if result:
        table_exists = True
    cur.close()

    return table_exists



"""
This module contains utility functions for MySQL Databases.
"""
import os
import psycopg2

S3_ACCESS_KEY = os.getenv("S3_ACCESS_KEY")
S3_SECRET_KEY = os.getenv("S3_SECRET_KEY")


def get_connection(schema_name, host, port, user, password):
    """
    This function will establish a database connection.
    """
    return psycopg2.connect(
        dbname=schema_name, host=host, port=port, user=user,
        password=password)


def load_from_s3(
        conn, table_name, s3_location, separator=",", header=True,
        file_format="csv"):
    """
    Loads data from an S3 location into a specified table in a PostgreSQL
    database managed by psycopg2.

    Args:
        conn (psycopg2.extensions.connection): The database connection object.
        table_name (str): Name of the table to which data will be loaded.
        s3_location (str): S3 URI where the data files are stored.
        separator (str, optional): File separator.
          Defaults to ','.
        header (bool, optional): Whether files have a header line.
          Defaults to True.
        file_format (str, optional): Format of the files ("csv" or "parquet").
          Defaults to "csv".

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    # Setup conversion parameters for csv
    # TODO: implement header_info argument into redshift query format
    ignore_header = " IGNOREHEADER 1 " if header else " "
    conversion_params = (
        f"DELIMITER '{separator}'{ignore_header}MAXERROR AS 10000;")
    if file_format == "parquet":
        conversion_params = "FORMAT AS PARQUET;"
    # Fill parameters
    query = f"""
        COPY {table_name}
        FROM '{s3_location}'
        ACCESS_KEY_ID '{S3_ACCESS_KEY}'
        SECRET_ACCESS_KEY '{S3_SECRET_KEY}'
        {conversion_params}
    """
    # Execute query
    try:
        with conn.cursor() as cur:
            cur.execute(query)
        conn.commit()
        return True
    except Exception as e:
        # TODO: Improve logging here by standardizing query execution.
        print(e)
        return False



"""
This module contains all DataBricks interactive functionality.
"""
import os
import base64
import json
import requests
from databricks_cli.sdk.api_client import ApiClient
from databricks_cli.sdk.service import WorkspaceService
from databricks_cli.dbfs.cli import DbfsApi


def update_global_init_script(host, token, script_text, script_id, name):
    """
        This method will update a DataBricks global init script.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            script_text (str): init script plain text
            script_id (str): global init script ID
            name (str): name of init script

        Returns:
            request response
    """
    return requests.request(
        "PATCH",
        os.path.join(host, "api/2.0/global-init-scripts", script_id),
        data=json.dumps({
            "name": name, "script": base64.b64encode(
                bytes(script_text, "utf-8")).decode("ascii")}),
        headers={"Authorization": f"Bearer {token}"})


def update_job(host, token, job_id, **kwargs):
    """
        This method will run a DataBricks job given the job id.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            job_id (int): respective job id

        Returns:
            request response
    """
    return requests.post(
        os.path.join(host, "api/2.0/jobs/update"),
        headers={"Authorization": f"Bearer {token}"},
        json={"job_id": job_id, **kwargs})


def run_job(host, token, job_id, **kwargs):
    """
        This method will run a DataBricks job given the job id.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            job_id (int): respective job id

        Returns:
            request response
    """
    return requests.post(
        os.path.join(host, 'api/2.0/jobs/run-now'),
        headers={"Authorization": f"Bearer {token}"},
        json={"job_id": job_id, **kwargs})


def get_run_status(host, token, run_id):
    """
        Get DataBricks run status information given a run id.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            run_id (int): respective run id

        Returns:
            request response
    """
    return requests.get(
        os.path.join(host, 'api/2.0/jobs/runs/get'), json={"run_id": run_id},
        headers={"Authorization": f"Bearer {token}"})


def get_job_info(host, token, job_id):
    """
        Get DataBricks job information given a job id.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            run_id (int): respective run id

        Returns:
            request response
    """
    return requests.get(
        os.path.join(host, 'api/2.0/jobs/get'), json={"job_id": job_id},
        headers={"Authorization": f"Bearer {token}"})


def delete_job_run(host, token, run_id):
    """
        This method will delete a specified run of a job configuration.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            run_id (int): respective run id

        Returns:
            request response
    """
    return requests.post(
        os.path.join(host, "api/2.0/jobs/runs/delete"),
        headers={"Authorization": f"Bearer {token}"},
        json={"run_id": run_id})


def get_task_params(host, token, job_id):
    """
        This method will get the existing task parameters given a job id.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            job_id (int): respective job id

        Returns:
            task parameter list and success boolean
    """
    # Initialize values
    success = False
    task_params = []
    # Get all job information
    response = get_job_info(host, token, job_id)
    # If a successful response has been retrieved pull the task parameters
    if response.status_code == 200:
        success = True
        # Parse response
        content = response.json()
        # If there are multiple tasks build the list
        if "tasks" in content["settings"]:
            task_params = [
                {
                    key: value for key, value in i.items()
                    if key in ["task_key", "notebook_task"]
                }
                for i in content["settings"]["tasks"]]
        # Otherwise, use notebook_task field and fill task key with empty str
        else:
            task_params = [{
                "task_key": "",
                "notebook_task": content["settings"]["notebook_task"]}]

    return task_params, success


def export_notebook(host, token, notebook_path):
    """
        This method will export a DataBricks notebook to a python source file
        locally stored.

        Arguments:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            notebook_path (str): local DataBricks path to notebook

        Returns:
            source code string
    """
    # Initialize DataBricks API Client and Workspace API
    databricks_client = ApiClient(host=host, token=token, verify=True)
    workspace_service = WorkspaceService(databricks_client)
    # Export Workspace
    output = workspace_service.export_workspace(notebook_path, 'SOURCE')
    # Return decoded source code as a string
    return base64.b64decode(output["content"]).decode()


def import_notebook(host, token, source_code, notebook_path):
    """
        Import source code into DataBricks given a location.

        TODO: Add success boolean as return vaue

        Arguments:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            source_code (str): string of python source code
            notebook_path (str): location of notebook

        Returns:
            None
    """
    # Initialize DataBricks API Client and Workspace API
    databricks_client = ApiClient(host=host, token=token, verify=True)
    workspace_service = WorkspaceService(databricks_client)
    # Import the source code as the new DataBricks notebook
    workspace_service.import_workspace(
        notebook_path, "SOURCE", "PYTHON", source_code, True)

    return


def does_dir_exist(host, token, user, notebook_path):
    """
        This method will determine whether the directory in a notebook path
        exists.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            user (str): databricks user path
            notebook_path (str): location of notebook

        Returns:
            boolean
    """
    # Parse the directory name from the full notebook path
    dir_name = os.path.dirname(notebook_path.split(user)[1])
    # Initialize DataBricks API Client and Workspace API
    databricks_client = ApiClient(host=host, token=token, verify=True)
    workspace_service = WorkspaceService(databricks_client)
    # Grab the workspace objects for the provided user
    workspace_objects = workspace_service.list(user)
    # Return a boolean for whether the directory exists iin the workspace
    return any([
        user + dir_name == i["path"]
        for i in workspace_objects["objects"]
        if i["object_type"] == "DIRECTORY"])


def make_dir(host, token, user, notebook_path):
    """
        This method will make a databricks directory given user and notebook
        path information.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            user (str): databricks user path
            notebook_path (str): location of notebook

        Returns:

    """
    # Parse the directory name from the full notebook path
    dir_name = os.path.dirname(notebook_path.split(user)[1])
    # Initialize DataBricks API Client and Workspace API
    databricks_client = ApiClient(host=host, token=token, verify=True)
    workspace_service = WorkspaceService(databricks_client)
    # Make the directory
    return workspace_service.mkdirs(user + dir_name)


def put_file(host, token, src_path, dbfs_path):
    """
        Uploads a file from the local filesystem to Databricks File System.

        Args:
            host (str): URL of DataBricks workspace
            token (str): DataBricks API token
            src_path (str): The path of the file on the local filesystem.
            dbfs_path (str): DBFS path where the file will be uploaded.

        Returns:
            TODO: Add success boolean
    """
    # Connect to databricks
    databricks_client = ApiClient(host=host, token=token, verify=True)
    dbfs_api = DbfsApi(databricks_client)
    # Read file and upload to dbfs
    handle = dbfs_api.client.create(dbfs_path, True)['handle']
    with open(src_path, 'rb') as local_file:
        while True:
            contents = local_file.read(2 ** 20)
            if len(contents) == 0:
                break
            # add_block should not take a bytes object.
            dbfs_api.client.add_block(
                handle, base64.b64encode(contents).decode())
        dbfs_api.client.close(handle)

    return



"""
This module contains utility functions for MySQL Databases.
"""
import logging
import pymysql
from pymysql.constants import CLIENT


def get_connection(schema_name, host, port, user, password):
    """
    Establish a database connection using the provided parameters.

    This function attempts to connect to a MySQL database using the provided
    schema name, host, port, username, and password. It returns the connection
    object if successful; otherwise, it returns None and prints an error
    message.

    Args:
        schema_name (str): The name of the database schema.
        host (str): The hostname of the database server.
        port (int): The port number to connect to the database server.
        user (str): The username to connect to the database.
        password (str): The password for the provided username.

    Returns:
        pymysql.connections.Connection or None: The database connection object
        if successful, or None otherwise.

    Raises:
        This function will catch and handle all exceptions internally,
        therefore it does not raise exceptions itself.

    Examples:
        >>> conn = get_connection(
        >>>     "my_schema", "localhost", 3306, "user", "password")
        >>> type(conn)
        <class 'pymysql.connections.Connection'>
    """
    try:
        connection = pymysql.connect(
            db=schema_name, host=host, port=port, user=user, passwd=password,
            client_flag=CLIENT.MULTI_STATEMENTS)
        return connection
    # Handle operational errors like connection failure, etc.
    except pymysql.OperationalError as oe:
        print(f"Operational Error: {oe}")
        return None
    # Handle programming errors like database not found, etc.
    except pymysql.ProgrammingError as pe:
        print(f"Programming Error: {pe}")
        return None
    # General exception to catch any other errors
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return None


def load_from_s3(
        conn, table_name, s3_prefix, separator=",", header=True,
        replace=False, header_list=None):
    """
    Load data from an AWS S3 bucket into a MySQL database table.

    This function attempts to load data from a specified S3 prefix into a
    MySQL database table. The S3 dataset is expected to be in CSV format. 
    The function will commit the transaction if successful and rollback in
    case of failure. Errors that occur are logged.

    Args:
        conn (pymysql.connections.Connection): A valid database connection.
        table_name (str): The name of the target database table.
        s3_prefix (str): The S3 prefix where the dataset is stored.
        separator (str, optional): 
            The separator used in the CSV file. 
            Defaults to ','.
        header (bool, optional): 
            Indicates whether the first row of the CSV is a header. 
            Defaults to True.
        replace (bool, optional): 
            If True, existing records in the table will be replaced. 
            Defaults to False.
        header_list (list, optional): 
            An ordered list of header names that match the table columns. 
            Only needed if 'header' is False. 
            Defaults to None.

    Returns:
        bool: True if data loading is successful, otherwise False.

    Note:
        This function is designed to catch and log all exceptions, therefore
        it does not raise exceptions itself.

    Examples:
        >>> conn = pymysql.connect(...)  # Establish a database connection first
        >>> load_from_s3(
        >>>     conn, 'my_table', 's3://my-bucket/data', separator=';',
        >>>     header=True)
        True
    """
    # Initialize logging
    logging.basicConfig(level=logging.INFO)
    # Set ignore row string given header boolean
    ignore_rows = "\n    IGNORE 1 ROWS" if header else ""
    replace_rows = "REPLACE " if replace else ""
    # Setup list of headers if applicable
    header_info = ""
    if header_list:
        header_info = " " + str(tuple(f"`{i}`" for i in header_list))
    # Fill parameters
    query = f"""
        LOAD DATA FROM S3 PREFIX '{s3_prefix}'
        {replace_rows}INTO TABLE {table_name}
        FIELDS TERMINATED BY '{separator}'{ignore_rows}{header_info};
    """
    # Execute query
    try:
        with conn.cursor() as cur:
            cur.execute(query)
        conn.commit()
        logging.info(f"Data successfully loaded into {table_name} from {s3_prefix}.")
        return True
    except pymysql.MySQLError as e:
        logging.error(f"MySQL Error: {e}")
        conn.rollback()
    except Exception as e:
        logging.error(f"An unexpected error occurred: {e}")
        conn.rollback()

    return False



"""
Apache Spark Utility Methods
"""
import os
import logging
import pyspark.sql.types as ps_types
import pyspark.sql.functions as psf
from . import s3_utils

# Setup s3 keys
S3_ACCESS_KEY = os.getenv('S3_ACCESS_KEY')
S3_SECRET_KEY = os.getenv('S3_SECRET_KEY')


def create_spark_schema(column_headers, column_types):
    """
    This method will create the pyspark schema provited types and
    column headers.

    Returns:
        pyspark schema
    """
    columns = zip(column_headers, column_types)
    schema = ps_types.StructType([
        ps_types.StructField(
            col_name,
            getattr(ps_types, col_type)(),
            True)
        for (col_name, col_type) in columns])

    return schema


def rename_cols(spark_df, column_map):
    """
    This method will rename a spark dataframes columns provided a column map.

    Args:
        spark_df (pyspark.sql.dataframe.DataFrame): data
        column_map (dict): rename column mapping

    Returns:
        modified spark dataframe with new column headers
    """
    for old_col_header, new_col_header in column_map.items():
        spark_df = spark_df.withColumnRenamed(
            old_col_header, new_col_header)

    return spark_df


def get_equivalent_spark_type(pandas_type):
    """
    This method will retrieve the corresponding spark type given a pandas
    type.

    Source: https://stackoverflow.com/questions/37513355

    Args:
        pandas_type (str): pandas data type

    Returns:
        spark data type
    """
    type_map = {
        'datetime64[ns]': ps_types.TimestampType(),
        'int64': ps_types.LongType(),
        'int32': ps_types.IntegerType(),
        'float64': ps_types.DoubleType(),
        'float32': ps_types.FloatType()}
    if pandas_type not in type_map:
        return ps_types.StringType()

    return type_map[pandas_type]


def pandas_to_spark(spark, pandas_df):
    """
    This method will return a spark dataframe given a pandas dataframe.

    Args:
        spark (pyspark.sql.session.SparkSession): pyspark session
        pandas_df (pandas.core.frame.DataFrame): pandas DataFrame

    Returns:
        equivalent spark DataFrame
    """
    columns = list(pandas_df.columns)
    types = list(pandas_df.dtypes)
    p_schema = ps_types.StructType([
        ps_types.StructField(column, get_equivalent_spark_type(pandas_type))
        for column, pandas_type in zip(columns, types)])

    return spark.createDataFrame(pandas_df, p_schema)


def get_distinct_values(spark_df, column_header):
    """
    Get the list of distinct values within a DataFrame column.

    Args:
        spark_df (pyspark.sql.dataframe.DataFrame): data table
        column_header (str): header string for desired column

    Returns:
        list of distinct values from the column
    """
    distinct_values = spark_df.select(column_header).distinct().rdd.flatMap(
        lambda x: x).collect()

    return distinct_values


def get_distinct_values_sql(sql_context, table_name, column_header):
    """
    Get the list of distinct values within a SQL table.

    Args:
        sql_context (pyspark.sql.context.SQLContext): spark SQL context
        table_name (str): table name of column
        column_header (str): header string for desired column

    Returns:
        list of distinct values from the column
    """
    query = f"SELECT DISTINCT {column_header} FROM {table_name}"
    distinct_values = sql_context.sql(query).rdd.flatMap(
        lambda x: x).collect()

    return distinct_values


def convert_timestamp(
        df, column_header, timezone="UTC", new_column_header=None):
    """
    This method will convert an integer column to a timestamp.

    Args:
        df (pyspark.sql.dataframe.DataFrame): dataframe object
        column_header (str): name of column header
        timezone (str): which timezone to convert to
        new_column_header (str): optional new column header for timestamp

    Returns:
        modified DataFrame with converted timestamp column
    """
    if not new_column_header:
        new_column_header = column_header
    # Convert milliseconds to timestamp column
    df = df.withColumn(
        new_column_header,
        psf.to_utc_timestamp(
            psf.from_unixtime(
                psf.col(column_header) / 1000, 'yyyy-MM-dd HH:mm:ss'),
            timezone))

    return df


def convert_to_pandas(spark_df):
    """
    This function will safely convert a spark DataFrame to a pandas DataFrame.

    Source: https://stackoverflow.com/questions/76072664

    Args:
        spark_df (pyspark.sql.dataframe.DataFrame): spark DataFrame

    Returns:
        pandas DataFrame
    """
    # Iterate over columns and convert each timestamp column to a string
    timestamp_cols = []
    for column in spark_df.schema:
        if column.dataType == ps_types.TimestampType():
            # Append column header to list
            timestamp_cols.append(column.name)
            # Set column to string using date_format function
            spark_df = spark_df.withColumn(
                column.name,
                psf.date_format(column.name, "yyyy-MM-dd HH:mm:ss"))
    # Convert to a pandas DataFrame and reset timestamp columns
    pandas_df = spark_df.toPandas()
    for column_header in timestamp_cols:
        pandas_df[column_header] = pandas_df[
            column_header].astype("datetime64[ns]")

    return pandas_df


def save_spark_df_to_s3(
        spark_df, output_location, file_format="csv", separator=",",
        use_pandas=False, header=True, partition_by=[], repartition={},
        replace_where=[], mode="overwrite", max_records_per_file=None):
    """
    This method will save a Spark DataFrame to a provided s3 location.

    TODO: Break this out and simplify it.

    Args:
        spark_df (pyspark.sql.dataframe.DataFrame):
        output_location (str): output s3 location
        file_format (str): file format to save dataframe to
        separator (str): column delimiter
        use_pandas (bool): whether to convert to pandas and save

    Returns:
        None
    """
    # If the spark DataFrame is small then convert to pandas
    if use_pandas:
        # Convert to pandas
        pandas_df = convert_to_pandas(spark_df)
        # Write result to specified s3 location
        s3_utils.write_pandas_df(
            S3_ACCESS_KEY, S3_SECRET_KEY, output_location, pandas_df,
            file_format=file_format, sep=separator, header=header)
        # Log save information
        logging.info("{row_count} rows unloaded to {output_location}".format(
            row_count=pandas_df.shape[0], output_location=output_location))
    # Otherwise, use spark to write
    else:
        # Repartition result DataFrame provided arguments
        if repartition:
            args = []
            # Add number of partitions first if applicable
            if "n_partitions" in repartition:
                args.append(repartition["n_partitions"])
            # Add columns headers if applicable
            if "column_headers" in repartition:
                args += [psf.col(c) for c in repartition["column_headers"]]
            # Repartition
            spark_df = spark_df.repartition(*args)
        # Setup options
        options_dict = {}
        # Setup max records if provided
        if max_records_per_file:
            options_dict["maxRecordsPerFile"] = max_records_per_file
        # Setup general parameters based on output file format
        if file_format in ('parquet', 'delta'):
            options_dict["mergeSchema"] = "true"
            if (
                file_format == 'delta' and
                replace_where and
                mode == 'overwrite'
            ):
                options_dict["replaceWhere"] = " AND ".join(replace_where)
        elif file_format == 'csv':
            options_dict["delimiter"] = separator
            options_dict["header"] = str(header).lower()
        # Write data to s3
        if partition_by:
            spark_df \
                .write \
                .format(file_format) \
                .partitionBy(partition_by) \
                .mode(mode) \
                .options(**options_dict) \
                .save(output_location)
        else:
            spark_df \
                .write \
                .format(file_format) \
                .mode(mode) \
                .options(**options_dict) \
                .save(output_location)
        # Log output location
        logging.info(f"Data unloaded to {output_location}")

    return



"""
AWS S3 Blob Storage Utility Methods
"""
import io
import logging
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor
import zipfile
import yaml
import boto3
import numpy as np
import pandas as pd

# Setup logging
logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s:%(message)s")
logging.getLogger("py4j").setLevel(logging.ERROR)

# https://stackoverflow.com/questions/51272814
yaml.Dumper.ignore_aliases = lambda *args: True


def is_valid_s3_url(s3_url):
    """
    Check if the given URL is a valid S3 URL.

    Args:
        s3_url (str): The URL to check.

    Returns:
        bool: True if valid, False otherwise.
    """
    try:
        parsed = urlparse(s3_url)
    except ValueError:
        return False
    
    return parsed.scheme == 's3' and bool(parsed.netloc) and bool(parsed.path)


def parse_url(s3_url):
    """
        This method will parse an s3 url.

        Args:
            s3_url (str): s3 url

        Returns:
            prefix and bucket name
    """
    if not is_valid_s3_url(s3_url):
        return None, None
    # Parse proper output url
    parse_result = urlparse(s3_url)
    # Return prefix and bucket name
    return parse_result.path[1:], parse_result.netloc


def read_file(access_key, secret_key, s3_prefix, bucket_name):
    """
        This method will read files from s3 using a boto3 client.

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            s3_prefix (str): AWS s3 prefix to file
            bucket_name (str): AWS s3 bucket name

        Returns:
            bytes object
    """
    client = boto3.client(
        's3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
    file = client.get_object(Bucket=bucket_name, Key=s3_prefix)

    return file['Body'].read()


def read_df(access_key, secret_key, s3_prefix, bucket_name, **kwargs):
    """
        This method will read in data from s3 into a pandas DataFrame.
        TODO: Add support for parquet

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            s3_prefix (str): AWS s3 prefix to file
            bucket_name (str): AWS s3 bucket name

        Returns:
            bytes object
    """
    return pd.read_csv(
        io.StringIO(str(read_file(
            access_key, secret_key, s3_prefix, bucket_name), "utf-8")),
        # Pass additional keyword arguments to pandas read_csv method
        **kwargs)


def write_bytes(access_key, secret_key, s3_prefix, bucket_name, bytes_object):
    """
        This method will write a bytes object to s3 provided a prefix.

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            s3_prefix (str): AWS s3 prefix to file
            bucket_name (str): AWS s3 bucket name
            bytes_object (bytes): object that will be written

        Returns:
            Success boolean
    """
    # Setup boto3 s3 client
    client = boto3.client(
        's3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
    # Write object to s3
    response = client.put_object(
        Body=bytes_object, Bucket=bucket_name, Key=s3_prefix)
    # Return success
    return response["ResponseMetadata"]["HTTPStatusCode"] == 200


def write_pandas_df(
        access_key, secret_key, s3_url, pandas_df, file_format="csv",
        **kwargs):
    """
    This method will save a DataFrame to S3 provided the filename.

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        s3_url (str): s3 url where data will be written
        pandas_df (pandas.DataFrame): DataFrame object
        file_format (str): desired file format, default is csv

    Returns:
        success boolean
    """
    # Encode pandas DataFrame to bytes object
    if file_format == "csv":
        bytes_object = pandas_df.to_csv(None, index=False, **kwargs).encode()
    elif file_format == "parquet":
        bytes_object = pandas_df.to_parquet(None, index=False)
    else:
        return False
    # Write bytes to s3
    return write_bytes(
        access_key, secret_key,
        # Parse s3 URL for bucket name and s3 key
        *parse_url(s3_url),
        bytes_object)


def write_dict(access_key, secret_key, s3_url, dict_object):
    """
    This method will convert a dict to bytes using YAML and write them to
    a specified s3 location.

    TODO: Add other file format options like JSON

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        s3_url (str): s3 url where data will be written
        dict_object (dict): python dictionary

    Returns:
        success boolean
    """
    return write_bytes(
        access_key, secret_key,
        # Parse s3 URL for bucket name and s3 key
        *parse_url(s3_url),
        # Encode dictionary
        yaml.dump(dict_object).encode())


def write_zip(access_key, secret_key, s3_url, file_dict):
    """
    This method will zip a dictionary of byte objects and save the file
    on s3.

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        s3_url (str): s3 url where data will be written
        file_dict (dict): filenames and their corresponding bytes

    Returns:
        success boolean
    """
    # Write bytes in memory
    zip_buffer = io.BytesIO()
    with zipfile.ZipFile(
        zip_buffer, "a", zipfile.ZIP_DEFLATED, allowZip64=True
    ) as zip_file:
        for key, value in file_dict.items():
            zip_file.writestr(key, value)
    # Write bytes buffer to file
    success = write_bytes(
        access_key, secret_key, *parse_url(s3_url), zip_buffer.getvalue())
    # Close buffer
    zip_buffer.close()

    return success


def write_local_file(
        access_key, secret_key, s3_prefix, bucket_name, local_filepath):
    """
    This method will write a local file to s3.

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        s3_prefix (str): AWS s3 prefix to file
        bucket_name (str): AWS s3 bucket name
        local_filepath (str): local filepath

    Returns:
        success
    """
    success = False
    with open(local_filepath, "rb") as bytes_object:
        success = write_bytes(
            access_key, secret_key, s3_prefix, bucket_name, bytes_object)

    return success


def check_s3_path(access_key, secret_key, s3_path, bucket_name):
    """
    This method will check whether the provided s3 path is valid.

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        s3_path (str): path to s3 file
        bucket_name (str): name of s3 bucket

    Returns:
        boolean for whether the path exists
    """
    s3_client = boto3.client(
        's3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
    # --- Setup key ---
    # Remove bucket from path to get prefix if applicable
    if bucket_name in s3_path:
        s3_prefix = s3_path.split(bucket_name)[1][1:]
    else:
        s3_prefix = s3_path
    # Get prefix to the left of the glob character
    if "*" in s3_prefix:
        s3_prefix = s3_prefix.split("*")[0]
    # Get list response
    resp = s3_client.list_objects(
        Bucket=bucket_name, Prefix=s3_prefix, MaxKeys=1)

    return "Contents" in resp


def get_responses(
        access_key, secret_key, s3_prefix, bucket_name):
    """
        This method will get the file information for a given directory on s3.

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            s3_prefix (str): directory within s3 bucket
            bucket_name (str): name of s3 bucket

        Returns:
            list of json responses from S3
    """
    client = boto3.session.Session().client(
        "s3", aws_access_key_id=access_key, aws_secret_access_key=secret_key)
    continuation_token = None
    responses = []
    # List objects within the given directory until the response is truncated
    while True:
        list_kwargs = dict(Bucket=bucket_name, Prefix=s3_prefix, MaxKeys=1000)
        # Add continuation token if not None
        if continuation_token:
            list_kwargs['ContinuationToken'] = continuation_token
        response = client.list_objects_v2(**list_kwargs)
        # Add valid reponses and update continuation token
        if 'Contents' in response:
            responses += response['Contents']
        # Exit while loop if at the end of the objects
        if not response.get('IsTruncated'):
            break
        continuation_token = response.get('NextContinuationToken')

    return responses


def get_s3_prefix_size(access_key, secret_key, s3_prefix_list, bucket_name):
    """
    This method will get the size of a list of s3 prefixes.

    Args:
        access_key (str): AWS s3 Access Key
        secret_key (str): AWS s3 Secret Key
        prefix_list (list): list of s3 prefixes
        bucket (str): name of s3 bucket

    Returns:
        size (in Gb)
    """
    responses = []
    for s3_prefix in s3_prefix_list:
        responses += get_responses(
            access_key, secret_key, s3_prefix, bucket_name)
    return round(sum([i["Size"] for i in responses]) / np.power(10, 9), 2)


def copy_file(access_key, secret_key, old_prefix, new_prefix, bucket_name):
    """
        This method will copy a file in s3 given a task definition.

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            old_prefix (str): old s3 prefix
            new_prefix (str): new s3 prefix
            bucket_name (str): s3 bucket name

        Returns:
            success boolean and exception message
    """
    success = True
    s3_client = boto3.resource(
        's3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
    # Try to copy file
    try:
        s3_client.Object(bucket_name, new_prefix).copy_from(
            CopySource=f"{bucket_name}/" + old_prefix)
    # If the copy fails for any reason set success to False
    except Exception as e:
        success = False

    return success


def copy_s3_files(
        access_key, secret_key, key_map, bucket_name, worker_count=8,
        max_retries=1):
    """
        This method will function as a threaded s3 copy a set of key value
        pairs of old s3 keys and new s3 keys.

        Args:
            access_key (str): AWS s3 Access Key
            secret_key (str): AWS s3 Secret Key
            key_map (dict): map of old to new s3 prefixes
            bucket_name (str): s3 bucket
            worker_count (int): number of workers to spin up for copying
            max_retries (int): maximum number of copy retries for failed tasks

        Returns:
            success boolean
    """
    success = True
    old_keys = list(key_map.keys())
    new_keys = list(key_map.values())
    # Create a thread pood to multiprocess tasks
    with ThreadPoolExecutor(max_workers=worker_count) as executor:
        # Run initial tasks
        file_count = len(old_keys)
        results = list(executor.map(
            copy_file, [access_key] * file_count, [secret_key] * file_count,
            old_keys, new_keys, [bucket_name] * file_count))
        failed_tasks = [i for i in range(len(results)) if not results[i]]
        # Retry at most the number of max retries
        retries = 0
        while len(failed_tasks) > 0 and retries < max_retries:
            # Resubmit failed tasks
            futures = [i.result() for i in [
                executor.submit(
                    copy_file, access_key, secret_key, old_keys[i],
                    new_keys[i], bucket_name)
                for i in failed_tasks]]
            # Get updated failed tasks
            failed_tasks = [
                i for i in range(len(futures)) if not futures[i]]
            # Iterate retries
            retries += 1
        # If there are still failed tasks log error and reset success
        copies_failed = len(failed_tasks)
        if copies_failed > 0:
            logging.error(
                f"WARNING: {copies_failed} copies failed.")
            success = False

    return success



"""
This class will function as a general purpose DataBase API.
"""
import logging
from marshmallow import Schema, fields, post_load, validates, ValidationError
from .utilities import mysql_utils, postgresql_utils


class DatabaseSchema(Schema):
    """
    Schema for specifying database specs.
    """
    database_name = fields.String(required=True)
    database_type = fields.String(required=True)
    host = fields.String(required=True)
    port = fields.Integer(required=True)
    user = fields.String(required=True)
    password = fields.String(required=True)

    @validates("database_type")
    def validate_database_type(self, database_type):
        """ This function will validate the database type """
        valid_args = ["postgresql", "mysql"]
        if database_type not in valid_args:
            raise ValidationError(
                f"Invalid database_type '{database_type}' provided, "
                "please choose among the list: [{}]".format(
                    ", ".join(valid_args)))

    @post_load
    def create_database(self, input_data, **kwargs):
        return Database(**input_data)


class Database:
    """
    This class will provide an generic interface layer on top of our database.
    """

    def __init__(
            self, database_name, database_type, host, port, user, password):
        """
        Setup database interface arguments.
        """
        self.database_name = database_name
        self.database_type = database_type
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def get_connection(self, schema_name):
        """
        This wrapper function will get a database connection.
        """
        if self.database_type == "mysql":
            return mysql_utils.get_connection(
                schema_name, self.host, self.port, self.user, self.password)
        return postgresql_utils.get_connection(
            schema_name, self.host, self.port, self.user, self.password)

    def delete(
            self, schema_name, table_name, delete_all=False, days=None,
            column_header=None):
        """
        This method will delete and optimize a table provided inputs.

        Args:
            schema_name (str): name of database schema
            table_name (str): name of database table
            delete_all (bool): whether to delete all rows
            days (int): number of days out to delete
            column_header (str): header for day filter

        Returns:
            deletion success boolean
        """
        delete_success = False
        # Connect to database
        conn = self.get_connection(schema_name)
        # Construct delete query string
        delete_query = f"DELETE FROM {table_name}"
        if delete_all:
            delete_query += ";"
        elif ((days is not None) and (column_header is not None)):
            delete_query += (
                f" WHERE DATE({column_header}) <= "
                f"CURDATE() - INTERVAL {days} DAY;")
        else:
            logging.error(
                "Either provide delete_all True or both days and column_header")
            return delete_success
        # Add table optimization if load location is Aurora
        if self.database_type == "mysql":
            delete_query += f"\nOPTIMIZE TABLE {table_name};"
        try:
            with conn.cursor() as cur:
                cur.execute(delete_query)
            conn.commit()
            delete_success = True
            logging.info(
                f"Successfully deleted rows beyond {days} days from {table_name}")
        except Exception as e:
            logging.error(f"Failed table deletion: {e}")

        return delete_success

    def load_into(self, schema_name, table_name, s3_location, **kwargs):
        """
        This wrapper function will load data into the database from S3.
        """
        # Connect to database
        conn = self.get_connection(schema_name)
        # TODO: Add connection check here
        # Wrap load data method
        if self.database_type == "mysql":
            success = mysql_utils.load_from_s3(
                conn, table_name, s3_location, **{
                    key: value for key, value in kwargs.items()
                    if key in (
                        "separator", "header", "replace", "header_list")})
        else:
            success = postgresql_utils.load_from_s3(
                conn, table_name, s3_location, **{
                    key: value for key, value in kwargs.items()
                    if key in ("separator", "header", "file_format")})
        # Log either success or failure
        if success:
            logging.info(
                f"Successfully loaded into the {self.database_name} "
                f"table {schema_name}.{table_name}")
        else:
            logging.error(
                f"Failed loading into the {self.database_name} "
                f"table {schema_name}.{table_name}")
        # Close connection
        conn.close()

        return success



from github import Github


def get_repo_and_contents(host, token, repo_name, branch, repo_location):
    """
        This method will get the GitHub contents object provided arguments.

        Arguments:
            host (str): GitHub URL
            token (str): GitHub API token
            repo_name (str): name of GitHub repository
            branch (str): name of GitHub branch
            source_code (str): source code
            repo_location (str): local path to source code location in repo

        Returns:
            github ContentFile object
    """
    # Initialize GitHub instance using GHE Token
    github_instance = Github(base_url=host, login_or_token=token)
    # Return repo and contents objects
    repo = github_instance.get_repo(repo_name)
    return repo, repo.get_contents(repo_location, ref=branch)


def update_source_code(repo, contents, branch, source_code):
    """
        This method will update source code in GitHub using local source code.

        Arguments:
            repo (github.Repository.Repository): GitHub repository instance
            contents (github.ContentFile.ContentFile): file contents
            branch (str): name of GitHub branch
            source_code (str): source code

        Returns:
            commit dict
    """
    # Update source code
    return repo.update_file(
        contents.path, "Automated code update.", source_code, contents.sha,
        branch=branch)



"""
This module will house all automated slack app logic.
"""
import io
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError


def post_dataframe_as_message(channel, token, pandas_df, message):
    """
        Posts a pandas DataFrame to a Slack channel.

        Args:
            channel (str): The ID or name of the Slack channel to post to.
            token (str): The Slack API token to use for authentication.
            pandas_df (pd.DataFrame): The DataFrame to post to the channel.
            message (str): A message to include with the DataFrame.

        Returns:
            None.
    """
    post_text = (
        # Place the message above the DataFrame
        message + "\n\n" +
        # Convert the DataFrame to a string surrounded by backticks
        # Source for formatting: https://stackoverflow.com/questions/66713432
        "``` \n" + pandas_df.to_markdown(index=False, floatfmt='') + "\n ```")
    # Initialize a Slack WebClient instance
    client = WebClient(token=token)
    # Post the message to Slack
    try:
        response = client.chat_postMessage(channel=channel, text=post_text)
        print("Message posted to Slack")
    except SlackApiError as e:
        print("Error posting message to Slack: {}".format(e))


def post_dataframe_as_csv(channel_id, token, pandas_df, filename):
    """
        Posts a pandas DataFrame to a Slack channel.

        Args:
            channel_id (str): The ID or name of the Slack channel to post to.
            token (str): The Slack API token to use for authentication.
            pandas_df (pd.DataFrame): The DataFrame to post to the channel.
            filename (str): Filename for downloadable CSV.

        Returns:
            None.
    """
    # Write the DataFrame to a CSV in a buffer
    csv_buffer = io.StringIO()
    pandas_df.to_csv(csv_buffer, index=False)
    # Initialize a Slack WebClient instance
    client = WebClient(token=token)
    # Upload the CSV to a Slack channel
    try:
        response = client.files_upload_v2(
            channel=channel_id, filename=filename,
            content=csv_buffer.getvalue(),
            initial_comment="*~~~ CSV DOWNLOAD ~~~*")
        print("CSV posted to Slack")
    except SlackApiError as e:
        print("Error posting CSV to Slack: {}".format(e))



"""
Utility metohds that aren't directly related to some sort of software package
or service.
"""
import tarfile
import io
import re
import datetime
import itertools
import logging
import math
import numpy as np
from scipy.stats import shapiro, normaltest


def get_date_range(date_0, date_1):
    """
    This method creates a list of dates from d0 to d1.

    Args:
        date_0 (datetime.date): start date
        date_1 (datetime.date): end date

    Returns:
        date range
    """
    return [
        date_0 + datetime.timedelta(days=i)
        for i in range((date_1 - date_0).days + 1)]


def leading_zero(hour):
    """
    This method will generate an hour str with a leading zero.

    Args:
        hour (int|str): hour

    Returns:
        leading zero digit str
    """
    str_hour = str(hour)
    if ((len(str_hour) == 1) and (str_hour != "*")):
        str_hour = "0" + str_hour

    return str_hour


def get_dt_range(dt, days=1, hours=0, weeks=0):
    """
    This method will generate the datetime range between the provided datetime
    and days / hours. Subtract hours / days when negative, otherwise add
    them.

    Args:
        dt (datetime.datetime): input datetime
        days (int): number of days
        hours (int): number of hours

    Returns:
        date range list
    """
    # If both values passed in are 0 return input datetime
    if (days == 0) and (hours == 0) and (weeks == 0):
        logging.error("Both days, hours, and weeks arguments were all 0.")
        return [dt]
    # If hour passed in is the glob character adjust accordingly
    if hours == "*":
        # If days are negative set the dt hour to 23
        if days < 0:
            dt = datetime.datetime(dt.year, dt.month, dt.day, 23)
        # Set hours to 0
        hours = 0
        # Set days to 1 if 0 was passed in
        days = (1 if days == 0 else days)
    # If weeks is setup grab trailing weeks
    if weeks != 0:
        dt_range = [
            dt - datetime.timedelta(weeks=minus_week)
            for minus_week in range(1, weeks + 1)]
    else:
        # Combine days and hours
        total_hours = (24 * days) + hours
        # Get dt range
        if total_hours < 0:
            dt_range = [
                dt - datetime.timedelta(hours=hour)
                for hour in range(abs(total_hours))]
        else:
            dt_range = [
                dt + datetime.timedelta(hours=hour)
                for hour in range(total_hours)]

    return dt_range


def exclude_hours_from_range(dt_range, exclude_hours):
    """
    This method will take a list of strings, convert them to a list of
    integers, then remove all datetimes from the range where the hour
    matches any in the list of integers.

    Args:
        dt_range (list): list of datetimes
        exclude_hours (list): list of strings

    Returns:
        modified list with hours excluded
    """
    hours = []
    # Assemble list of hours
    for arg in exclude_hours:
        # If the type is an int just add it to the list
        if isinstance(arg, int):
            hours.append(arg)
        # Otherwise check if the argument is a string
        elif isinstance(arg, str):
            # If the argument string of digits, cast to int first
            if re.compile(r"[0-9]+").fullmatch(arg):
                hours.append(int(arg))
            # If the argument is a range of digits parse the range and add it
            elif re.compile(r"[0-9]+-[0-9]+").fullmatch(arg):
                start, end = [int(i) for i in arg.split("-")]
                hours += list(range(start, end + 1))
    # Update the dt range to exclude these hours and return
    return [i for i in dt_range if i.hour not in hours]


def create_time_string(total_seconds):
    """
    This method will create a time string given total seconds.

    Args:
        total_seconds (int|str): total time in seconds

    Returns:
        time string
    """
    # If the input is a string cast to an int
    if (
        isinstance(total_seconds, str) and
        re.compile(r"-*[0-9]+").fullmatch(total_seconds)
    ):
        total_seconds = int(total_seconds)
    # If total_seconds is still not an int set it to zero
    if not isinstance(total_seconds, int):
        total_seconds = 0
    # Determine whether total seconds is negative
    is_negative = total_seconds < 0
    # Take the absolute value of total seconds
    total_seconds = abs(total_seconds)
    # Break down the total seconds into minutes and seconds
    minutes = math.floor(total_seconds / 60.0)
    seconds = total_seconds % 60
    # Format the final string
    minutes_str = "{minutes} minute{s}".format(
        minutes=minutes, s="s" * (minutes != 1))
    seconds_str = "{seconds} second{s}".format(
        seconds=seconds, s="s" * (seconds != 1))
    time_str = (
        "-" * is_negative +
        minutes_str * (minutes != 0) +
        " and " * (minutes != 0 and seconds != 0) +
        seconds_str * (not (seconds == 0 and minutes > 0)))

    return time_str


def apply_time_delta(dt, hour, time_delta):
    """
    This method will apply the time delta arguments to the dt and hour
    objects needed for formatting dataset paths.

    Args:
        dt (datetime.datetime): datetime object
        hour (str): either the hour or a glob of all hours
        time_delta (dict): dict with hours and days fields

    Returns:
        new modified datetime and hour objects
    """
    # If hour is glob only consider time delta day arguments
    if hour == "*":
        td = datetime.timedelta(
            **{k: v for k, v in time_delta.items() if k == "days"})
        new_dt = dt - td
        new_hour = hour
    # Otherwise cast hour to int, setup dt, and subtract full time delta
    else:
        td = datetime.timedelta(**time_delta)
        new_dt = datetime.datetime(dt.year, dt.month, dt.day, int(hour)) - td
        new_hour = new_dt.hour

    return new_dt, new_hour


def get_dict_permutations(raw_dict):
    """
    This method will take a raw dictionary and create all unique
    permutations of key value pairs.

    Source: https://codereview.stackexchange.com/questions/171173

    Args:
        raw_dict (dict): raw dictionary

    Returns:
        list of unique key value dict permutations
    """
    # Set default
    dict_permutations = [{}]
    # Check whether input is valid nonempty dictionary
    if isinstance(raw_dict, dict) and (len(raw_dict) > 0):
        # Make sure all values are lists
        dict_of_lists = {}
        for key, value in raw_dict.items():
            if not isinstance(value, list):
                dict_of_lists[key] = [value]
            else:
                dict_of_lists[key] = value
        # Create all unique permutations
        keys, values = zip(*dict_of_lists.items())
        dict_permutations = [
            dict(zip(keys, v)) for v in itertools.product(*values)]

    return dict_permutations


def pooled_stddev(stddevs, sample_sizes):
    """
    This method will calculate the pooled standard deviation across a
    group of samples given each samples standard deviation and size.

    Source: https://www.statisticshowto.com/pooled-standard-deviation/

    Args:
        stddevs (numpy.ndarray): standard deviations of samples
        sample_sizes (numpy.ndarray): samples sizes

    Returns:
        pooled stddev
    """
    # Return null if both arrays are empty or different lengths
    if (
        not(stddevs.any() and sample_sizes.any()) or
        (len(stddevs) != len(sample_sizes))
    ):
        return np.nan
    # Otherwise calculate the pooled standard deviation
    return np.sqrt(np.sum([
        (sample_sizes[i] - 1) * np.power(stddevs[i], 2)
        for i in range(len(sample_sizes))
    ]) / (np.sum(sample_sizes) - len(sample_sizes)))


def test_normal(values, alpha=0.05):
    """
    This method will test whether distributions are guassian.
    TODO: This needs to be flipped

    Args:
        values (np.array):

    Return:
        boolean result
    """
    _, shapiro_p = shapiro(values)
    _, normal_p = normaltest(values)

    return np.all([p < alpha for p in (shapiro_p, normal_p)])


def read_tar_from_bytes(tar_bytes):
    """
    Reads a tar archive from a bytes object and returns a dictionary.
    The dict keys are file names and values are the file content as strings.
    
    Parameters:
        tar_bytes (bytes): The bytes object containing the tar archive.
        
    Returns:
        dict: A dictionary with file names as keys and file content as values.
    """
    file_contents = {}
    # Create a BytesIO object from the bytes
    tar_stream = io.BytesIO(tar_bytes)
    # Open the tar file from the BytesIO stream
    with tarfile.open(fileobj=tar_stream) as tar:
        # Loop over each file in the archive
        for member in tar.getmembers():
            # Make sure it's actually a file (not a directory, etc.)
            if member.isfile():
                # Extract the file object
                f = tar.extractfile(member)
                # Read the file content and decode it to string
                file_contents[member.name] = f.read().decode("latin1")
                
    return file_contents


def write_dict_to_tar_bytes(file_dict):
    """
    Writes a dictionary to a tar archive as a bytes object.
    The dictionary keys are treated as file names, and the values as file content.
    
    Parameters:
        file_dict (dict): A dictionary with file names as keys and file content as values.
        
    Returns:
        bytes: The bytes object containing the tar archive.
    """
    tar_stream = io.BytesIO()
    # Create a tar archive in memory
    with tarfile.open(fileobj=tar_stream, mode='w:gz') as tar:
        for file_name, file_content in file_dict.items():
            # Convert string content to bytes
            file_data = file_content.encode("latin1")
            # Create a TarInfo object for the file
            file_info = tarfile.TarInfo(name=file_name)
            file_info.size = len(file_data)
            # Add the file to the tar archive
            tar.addfile(file_info, io.BytesIO(file_data))
    # Return the byte content of the tar archive
    return tar_stream.getvalue()



import re
import random


MAC_REGEX = re.compile(r"((?:[0-9A-Fa-f]{2}[:-]?){5}[0-9A-Fa-f]{2})")
LOCAL_MAC_REGEX = re.compile(
    # First octet's second least significant bit must be 1
    r"((?:[0-9a-f][13579BbDdFf][:-]?){1}"
    r"([0-9A-Fa-f]{2}[:-]?){4}[0-9A-Fa-f]{2})")
IPv4_REGEX = re.compile(
    r"(?<![.\w])"  # Negative lookbehind
    r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    r"\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    r"\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    r"\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    r"(?!\w)"  # Negative lookahead for only word characters
)
# Partial source: https://stackoverflow.com/questions/53497
IPv6_REGEX = re.compile(
    r"(?<![.\w])"  # Negative lookbehind
    r"("  
    r"(([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|"
    r"(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|"
    r"((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|"
    r"(([0-9A-Fa-f]{1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|"
    r":((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|"
    r"(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|"
    r"((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|"
    r"(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|"
    r"((:[0-9A-Fa-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|"
    r"(([0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|"
    r"((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|"
    r"(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{1,4}){1,6})|"
    r"((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|"
    r"(:(((:[0-9A-Fa-f]{1,4}){1,7})|"
    r"((:[0-9A-Fa-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)"
    r"(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))"
    r")"
    r"(?!\w)"  # Negative lookahead for only word characters
)


def add_colons_to_mac(mac):
    """Add colons to a MAC address string.

    Args:
        mac (str):
            A 12-character MAC address string without any separators.

    Returns:
        str:
            The MAC address string with colons added between every two
            characters.

    Raises:
        ValueError: If the length of the input MAC address is not 12.

    Examples:
        >>> add_colons_to_mac("0123456789AB")
        "01:23:45:67:89:AB"

        >>> add_colons_to_mac("A1B2C3D4E5F6")
        "A1:B2:C3:D4:E5:F6"
    """
    if len(mac) != 12:
        raise ValueError("Invalid MAC address length")
    
    return ':'.join(mac[i:i+2] for i in range(0, 12, 2))


def find_unique_macs(text, case=None):
    """
    Find the unique mac addresses within some text.

    Args:
        text (str): text string
        case (str): specify whether to cast macs to uppercase or lowercase

    Returns:
        list of unique mac addresses
    """
    # Search for all MAC addresses in the text
    mac_addresses = re.findall(MAC_REGEX, text)
    # Since re.findall() returns tuples, convert them back to the original
    # mac addresses
    mac_addresses = ["".join(mac) for mac in mac_addresses]
    # Add colons to mac addresses if applicable
    mac_addresses = [
        add_colons_to_mac(mac) if ":" not in mac else mac
        for mac in mac_addresses]
    # Cast to provided case if applicable
    if case == "upper":
        mac_addresses = [mac.upper() for mac in mac_addresses]
    elif case == "lower":
        mac_addresses = [mac.lower() for mac in mac_addresses]
    # Cast to a set in order to recude the list to unique macs
    unique_macs = list(set(mac_addresses))
    # Sort the list before returning it
    unique_macs.sort()

    return unique_macs


def generate_random_mac():
    """
    Generate a random mac address.

    Returns:
        random mac address
    """
    return ":".join("{:02x}".format(random.randint(0, 255)) for _ in range(6))


def generate_local_mac_address():
    """
    Generate a random local MAC address.

    The function generates a random MAC address and ensures that it is a local
    MAC address by setting the second least significant bit of the first octet
    to 1.

    Returns:
        str:
            A MAC address string in the format "XX:XX:XX:XX:XX:XX", where each
            "XX" is a two-digit hexadecimal number.

    Examples:
        >>> generate_local_mac_address()
        "01:23:45:67:89:AB"

        >>> generate_local_mac_address()
        "1A:2B:3C:4D:5E:6F"
    """
    # Local MAC addresses have the second least significant bit of the first
    # octet set to 1. To achieve that, we randomly generate the lower 7 bits
    # and then set the second least significant bit to 1
    first_octet = random.randint(0, 127) << 1 | 1
    mac_address = [first_octet] + [random.randint(0, 255) for _ in range(5)]
    return ':'.join(f'{octet:02x}' for octet in mac_address)


def redact_macs_from_text(text, mac_map=None, case=None):
    """
    Provided some text, redact the original macs.

    Args:
        text (str): text string
        mac_map (dict): key value pairs of og macs and random macs
        case (str): specify whether to cast macs to uppercase or lowercase

    Returns:
        redacted text and updated mac map
    """
    # Pull unique mac lists
    mac_list = find_unique_macs(text, case=case)
    # If existing map is passed update it
    if mac_map:
        for og_mac in mac_list:
            if og_mac not in mac_map:
                # If the mac is a local mac just map it to itself
                if LOCAL_MAC_REGEX.fullmatch(og_mac):
                    mac_map.update({og_mac: og_mac})
                # Otherwise map the op mac to a randomly generated local mac
                else:
                    mac_map.update({og_mac: generate_local_mac_address()})
    # Otherwise create map of original mac address to random mac address
    else:
        mac_map = {
            og_mac: og_mac if LOCAL_MAC_REGEX.fullmatch(og_mac)
            else generate_local_mac_address()
            for og_mac in mac_list}
    # Replace instances of macs in text
    redacted_text = text
    # Replace each original mac with a redacted mac
    for og_mac, redacted_mac in mac_map.items():
        # Replace uppercase
        redacted_text = redacted_text.replace(og_mac.upper(), redacted_mac)
        # Replace lowercase
        redacted_text = redacted_text.replace(og_mac.lower(), redacted_mac)

    return redacted_text, mac_map


def find_unique_ipv4(text):
    """
    Finds and returns the unique IPv4 addresses in a given text.
    
    Args:
        text (str): The text to search for IPv4 addresses.
        
    Returns:
        list: A sorted list of unique IPv4 addresses found in the text.
    """
    ipv4_addresses = re.findall(IPv4_REGEX, text)
    unique_ipv4_addresses = list(set(ipv4_addresses))
    unique_ipv4_addresses.sort()

    return unique_ipv4_addresses


def find_unique_ipv6(text, case=None):
    """
    Finds and returns the unique IPv6 addresses in a given text.
    
    Args:
        text (str): The text to search for IPv6 addresses.
        
    Returns:
        list: A sorted list of unique IPv6 addresses found in the text.
    """
    ipv6_addresses = [
        match[0] for match in re.findall(IPv6_REGEX, text)]
    if case == "upper":
        ipv6_addresses = [ipv6.upper() for ipv6 in ipv6_addresses]
    elif case == "lower":
        ipv6_addresses = [ipv6.lower() for ipv6 in ipv6_addresses]
    unique_ipv6_addresses = list(set(ipv6_addresses))
    unique_ipv6_addresses.sort()

    return unique_ipv6_addresses


def generate_random_ipv4():
    """
    Generates a random IPv4 address.
    
    Returns:
        str: A random IPv4 address.
    """
    return ".".join(str(random.randint(0, 255)) for _ in range(4))


def generate_random_ipv6():
    """
    Generates a random IPv6 address.
    
    Returns:
        str: A random IPv6 address.
    """
    return ":".join("{:x}".format(random.randint(0, 0xFFFF)) for _ in range(8))


def redact_ip_addresses_from_text(text, ip_address_map=None, case=None):
    """
    Provided some text, redact the original ip addresses.

    Args:
        text (str): text string
        ip_address_map (dict): key value pairs of og addresses and random ones
        case (str): specify whether to cast addresses to uppercase or lowercase

    Returns:
        redacted text and updated ip address map
    """
    # Pull unique mac lists
    ipv4_addresses = find_unique_ipv4(text)
    ipv6_addresses = find_unique_ipv6(text, case=case)
    # If existing map is passed update it
    if ip_address_map:
        # Update IPv4 Addresses
        ip_address_map.update({
            og_ip_address: generate_random_ipv4()
            for og_ip_address in ipv4_addresses
            if og_ip_address not in ip_address_map})
        # Update IPv6 Addresses
        ip_address_map.update({
            og_ip_address: generate_random_ipv6()
            for og_ip_address in ipv6_addresses
            if og_ip_address not in ip_address_map})
    # Otherwise create map of original mac address to random mac address
    else:
        ip_address_map = {
            og_ip_address: generate_random_ipv4()
            for og_ip_address in ipv4_addresses}
        ip_address_map.update({
            og_ip_address: generate_random_ipv6()
            for og_ip_address in ipv6_addresses})
    # Replace instances of macs in text
    redacted_text = text
    # Replace each original mac with a redacted mac
    for og_ip_address, redacted_ip_address in ip_address_map.items():
        # TODO: Handle mixed case in text
        # Replace uppercase
        redacted_text = redacted_text.replace(
            og_ip_address.upper(), redacted_ip_address)
        # Replace lowercase
        redacted_text = redacted_text.replace(
            og_ip_address.lower(), redacted_ip_address)

    return redacted_text, ip_address_map



"""
Pandas Utility methods.
"""
import itertools


def get_null_columns(pandas_df):
    """
    This function will get null columns.

    Args:
        pandas_df (pandas.core.frame.DataFrame): pandas DataFrame

    Returns:
        list of non null columns
    """
    return [
        column_header
        for column_header, is_null in pandas_df.isnull().all().items()
        if is_null]


def get_non_null_columns(pandas_df):
    """
    This function will get non null columns.

    Args:
        pandas_df (pandas.core.frame.DataFrame): pandas DataFrame

    Returns:
        list of non null columns
    """
    return [
        column_header
        for column_header, is_null in pandas_df.isnull().all().items()
        if not is_null]


def collapse_dataframe_columns(pandas_df):
    """
    This method will collapse DataFrame column values into a list.

    Args:
        pandas_df (pandas.DataFrame): pandas DataFrame

    Returns:
        list of unique column values
    """
    return list(set(itertools.chain.from_iterable([
        pandas_df[~pandas_df[col].isnull()][col].values.tolist()
        for col in pandas_df.columns])))


def filter_dataframe(
        pandas_df, cols, filter_out=False, use_substring=False,
        use_startswith=False):
    """
    This method will filter a DataFrame by a list of columns.

    Args:
        pandas_df (pandas.DataFrame): pandas DataFrame
        cols (list): list of desired columns
        filter_out (bool): switch to filter columns out of DataFrame
        use_substring (bool): switch to use substring logic
        use_startswith (bool): switch to use startswith logic

    Returns:
        filtered DataFrame
    """
    # Create condition lambda function
    if use_substring:
        column_filter = lambda c: any(
            str(substring) in c for substring in cols)
    elif use_startswith:
        column_filter = lambda c: any(
            c.startswith(substring) for substring in cols)
    else:
        column_filter = lambda c: c in cols
    # Return DataFrame with filtered columns
    if filter_out:
        return pandas_df.loc[
            :, [c for c in pandas_df.columns if not column_filter(c)]]

    return pandas_df.loc[
        :, [c for c in pandas_df.columns if column_filter(c)]]
